import { removeSensibleValues } from "remove-sensible-values";
import { Service } from "./service.mjs";
import { keyValue2Object } from "./util.mjs";

/**
 * Config providing service.
 * Dispatches config requests to services.
 * Also preserves them until a maching service becomes avaliable.
 * @property {Map<string,Object>} preservedConfigs values for services not already established
 */
export class ServiceConfig extends Service {
  /**
   * @return {string} 'config'
   */
  static get name() {
    return "config";
  }

  static get description() {
    return "Config providing service";
  }

  preservedConfigs = new Map();

  /**
   * Deliver configuration for a given service.
   * @param {string} name service name
   * @param {Object} config
   */
  async configFor(name, config) {
    this.trace(`configFor ${name}`);

    await this.start();

    const pc = this.preservedConfigs.get(name);
    if (pc !== undefined) {
      config = config === undefined ? pc : merge(config, pc);
      this.trace(`using preserved config ${name}`);
    }

    this.preservedConfigs.set(name, config);

    return config;
  }

  /**
   * Forget about preserved config of a service.
   * @param {string} name service name
   */
  clearPreserved(name) {
    this.preservedConfigs.delete(name);
  }

  /**
   * Set config entry.
   * @param {string} key path to the value
   * @param {any} value
   */
  async configureValue(key, value) {
    return this.configure(keyValue2Object(key, value));
  }

  /**
   *
   * @param {Array|Object} config
   */
  async configure(config) {
    if (config === undefined) {
      return;
    }

    const update = async (name, c) => {
      const s = this.owner.services[name];
      if (s === undefined) {
        delete c.name;

        const merged = merge(this.preservedConfigs.get(name), c);
        this.trace(
          `Preserve config ${name} ${JSON.stringify(
            removeSensibleValues(merged)
          )}`
        );
        this.preservedConfigs.set(name, merged);
      } else {
        return s.configure(c);
      }
    };

    await Promise.all(
      Array.isArray(config)
        ? config.map(c => update(c.name, c))
        : Object.entries(config).map(([k, v]) => update(k, v))
    );
  }

  /**
   * We always start immediate.
   * @return {boolean} true
   */
  get autostart() {
    return true;
  }

  get name() {
    return "config";
  }
}

/**
 * Merge from b into a.
 * When *a* and *b* are arrays of values only the none duplicates are appendend to *a*.
 * @param {any} a
 * @param {any} b
 * @return {any} merged b into a
 */
export function merge(a, b) {
  if (b === undefined || b === null) {
    return a;
  }

  if (Array.isArray(a)) {
    if (Array.isArray(b)) {
      return [...a, ...b.filter(x => !a.find(e => Object.is(e, x)))];
    }
    return [...a, b];
  }

  if (Array.isArray(b)) {
    return b;
  }

  if (b.constructor && b.constructor.name === "Buffer") {
    return b;
  }

  switch (typeof b) {
    case "function":
    case "string":
    case "number":
    case "boolean":
      return b;
    case "object":
      if (a === undefined || a === null || typeof a !== "object") {
        a = {};
      }
      Object.keys(b).forEach(k => (a[k] = merge(a[k], b[k])));
  }

  return a;
}
